// tfam_keyboard.c

#include "config.h"
#include "lib_mcu/timer/timer16_drv.h"
#include "hid_keys.h"
#include "lib_mcu/usb/usb_drv.h"
#include "usb_descriptors.h"
#include "modules/usb/device_chap9/usb_standard_request.h"
#include <avr/pgmspace.h>
#include <string.h>
#include <stdio.h>

extern U8 usb_configuration_nb;

// mapping of mac keyboard codes to USB HID key codes
// note that i have remapped:
//     enter    -> HID_ESCAPE
static U8 mac2hid[] PROGMEM = {
//             1,                3,             5,             7,            9,            B,                 D,              F,
           HID_A,            HID_S,         HID_D,         HID_F,        HID_H,        HID_G,             HID_Z,          HID_X, // 0
           HID_C,            HID_V,             0,         HID_B,        HID_Q,        HID_W,             HID_E,          HID_R, // 1
           HID_Y,            HID_T,         HID_1,         HID_2,        HID_3,        HID_4,             HID_6,          HID_5, // 2
      HID_EQUALS,            HID_9,         HID_7,     HID_MINUS,        HID_8,        HID_0, HID_RIGHT_BRACKET,          HID_O, // 3
           HID_U, HID_LEFT_BRACKET,         HID_I,         HID_P,   HID_RETURN,        HID_L,             HID_J, HID_APOSTROPHE, // 4
           HID_K,    HID_SEMICOLON, HID_BACKSLASH,     HID_COMMA,    HID_SLASH,        HID_N,             HID_M,     HID_PERIOD, // 5
         HID_TAB,     HID_SPACEBAR,  HID_BACKTICK, HID_BACKSPACE,   HID_ESCAPE,            0,                 0,   HID_LEFT_GUI, // 6
  HID_LEFT_SHIFT,    HID_CAPS_LOCK,  HID_LEFT_ALT,             0,            0,            0,                 0,              0, // 7

// keypad codes
               0,       HID_KP_DOT, HID_RIGHT_ARROW,           0,            0,            0,    HID_LEFT_ARROW,   HID_KP_CLEAR, // 0
  HID_DOWN_ARROW,                0,               0,           0, HID_KP_ENTER, HID_UP_ARROW,      HID_KP_MINUS,              0, // 1
               0,                0,        HID_KP_0,    HID_KP_1,     HID_KP_2,     HID_KP_3,          HID_KP_4,       HID_KP_5, // 2
        HID_KP_6,         HID_KP_7,               0,    HID_KP_8,     HID_KP_9,            0,               0,                0, // 3
};

/*
static U8 mac2ascii[] PROGMEM = {
// 1,   3,    5,   7,   9,   B,   D,    F,
  'A', 'S',  'D', 'F', 'H', 'G', 'Z',  'X', // 0
  'C', 'V',    0, 'B', 'Q', 'W', 'E',  'R', // 1
  'Y', 'T',  '1', '2', '3', '4', '6',  '5', // 2
  '=', '9',  '7', '-', '8', '0', ']',  'O', // 3
  'U', '[',  'I', 'P', 'r', 'L', 'J', '\'', // 4
  'K', ';', '\\', ',', '/', 'N', 'M',  '.', // 5
  't', ' ',  '`', 'b', 'e',   0,   0,  'c', // 6
  's', 'l',  'o',   0,   0,   0,   0,    0, // 7

// keypad codes
    0, '.',  '>',   0,   0,   0, '<',  'c', // 0
  'v',   0,    0,   0, 'e', '^', '-',    0, // 1
    0,   0,  '0', '1', '2', '3', '4',  '5', // 2
  '6', '7',    0, '8', '9',   0,   0,    0  // 3
};
*/

#define NULL_RESPONSE 0x7B
#define KEYPAD_RESPONSE 0x79
#define MODELNO_CMD 0x16
#define INQUIRY_CMD 0x10
#define INSTANT_CMD 0x14

static volatile bit kbd_connected; // whether or not the keyboard is responding to commands
static volatile U8 kbd_data; // buffer for data being sent to or received from the keyboard
static volatile U8 bit_count; // how many bits of kbd_data were sent/received [0,8]
static volatile bit keypad_mode; // whether the next response is coming from the numeric keypad
static volatile U8 key_states[12]; // 96 bits for each key (lower 64 are for the main keyboard, higer 32 are for the keypad)
static void (* volatile timer_callback)(); // function to call in the timer1 interrupt handler
static void (* volatile clock_callback)(); // function to call in the keyboard clock line interrupt handler
static U8 sent_report[8];

static void reset();
static void send_bit();
static void receive_bit();
static void set_receive_mode();
static void signal_ready_to_receive();
static void send_command(U8);
static void disable_clock_interrupt();
static void tfam_keyboard_populate_report(U8 *report);

void tfam_keyboard_task() {
    int i;
    bit send;
    U8 report[8];

    if (! Is_device_enumerated()) {
        return;
    }
   
    Usb_select_endpoint(EP_KBD_IN);
    if (! Is_usb_in_ready()) {
        return;
    }
    
    tfam_keyboard_populate_report(report);
    
    send = FALSE;
    for (i = 0; i < 8; i++) {
        if (sent_report[i] != report[i]) {
            send = TRUE;
            break;
        }
    }

    for (i = 0; i < 8; i++) {
        Usb_write_byte(report[i]);
    }
    Usb_ack_in_ready(); // tell the USB controller to flush the buffer

}


// populate a USB HID keyboard report with the data in the key_states array
static void tfam_keyboard_populate_report(U8 *report) {
    U8 key_code, key_val, major, mask, bit_block, report_idx;

    report[0] = 0;
    report[1] = 0;
    report_idx = 2;
    key_code = 0;
    for (major = 0; major < 12; major++) {
        bit_block = key_states[major];
        mask = 0x1;
        while (mask) {
            if (bit_block & mask) {
                key_val = pgm_read_byte(&mac2hid[key_code]);
                if (HID_IS_MODIFIER(key_val)) {
                    // modifier keys like shift and control are mapped to well-known bits
                    // in report[0]
                    report[0] |= 1 << (key_val & 0x7);
                } else if ((key_val != 0) && (report_idx < 8)) {
                    // all other keys get added to report[2..7]
                    report[report_idx++] = key_val;
                }
            }
            key_code++;
            mask <<= 1;
        }
    }
    while (report_idx < 8) {
        report[report_idx++] = 0;
    }
}

// zero out key_states
static void clear_key_states() {
    U8 i;
    
    for (i = 0; i < 12; i++) {
        key_states[i] = 0;
    }
}

// process the 8-bit response we just got from the keyboard, which is stored in kbd_data.
static void process_response() {
    bit down;
    U8 val, major, mask;

    if (kbd_data == NULL_RESPONSE) {
        // nothing to do
    } else if (kbd_data == KEYPAD_RESPONSE) {
        // this means that the next response will come from the keypad, which means that the
        // keycodes need to be interpreted differently.  there is no action to take right now
        // except to update the 'keypad_mode' flag, which happens at the end of the function
    } else if ((kbd_data & 0x1) == 0) {
        // error
    } else {
        // a key was pressed or released. update the corresponding bit in key_states
        down = (kbd_data & 0x80) ? FALSE : TRUE; 
        val = (kbd_data >> 1) & 0x3f;
        if (keypad_mode) {
            // numeric keypad codes 0-31 are mapped to key_state bits 64-95
            val |= 0x40;
        }
        major = val >> 3;
        if (major < 12) {
            mask = 1 << (val & 0x7);
            if (down) {
                key_states[major] |= mask;
            } else {
                key_states[major] &= (U8) ~mask;
            }
        }
        tfam_led_light();
    }
    keypad_mode = (kbd_data == KEYPAD_RESPONSE);
}

// OLD hardware (breadboard):
// D6/INT6: keyboard clock (always input)
// B3: keyboard data (bidirectional)

// NEW hardware (pcb):
// C7/INT4: clock
// C6: data


// initialize everything and start talking to the keyboard
void tfam_keyboard_task_init() {
    disable_clock_interrupt();
    Timer16_clear();
    timer_callback = NULL;
    kbd_connected = FALSE;
    kbd_data = 0;
    bit_count = 0;
    keypad_mode = FALSE;
    clear_key_states();
    sent_report[1] = 0xff; // this will force a report to be sent, since report[1] is always 0
    DDRC &= ~0x80; // set clock line to input mode
    PORTC |= 0x80; // enable pullup resistor on clock line to avoid noise when nothing is connected
    DDRC &= ~0x40; // set data line to input mode
    PORTC |= 0x40; // enable pullup resistor on data line
    Timer16_compare_a_it_enable();
    sei();
    reset();
}

// disable the clock line interrupt handler
static void disable_clock_interrupt() {
    clock_callback = NULL;
    EIMSK &= ~(1 << INT4); // disable clock interrupt
}

// enable the interrupt associated with the externally-driven keyboard clock line.
static void enable_clock_interrupt(bit falling, void (* callback)()) {
    EIMSK &= ~(1 << INT4); // disable clock interrupt
    EICRB |= (1 << ISC41);
    if (falling) {
        EICRB &= ~(1 << ISC40); // trigger interrupt on the falling edge of the clock signal
    } else {
        EICRB |= (1 << ISC40); // trigger interrupt on the rising edge of the clock signal
    }
    clock_callback = callback;
    if (callback != NULL) {
        EIMSK |= (1 << INT4); // enable clock interrupt
    }
}

// reset the state of the system and try to reset the keyboard
static void reset() {
    clock_callback = NULL;
    timer_callback = NULL;
    kbd_connected = FALSE;
    clear_key_states();
    send_command(MODELNO_CMD);
}

// start timer1, invoking 'callback' after 'time' milliseconds (if 'ms' is true) or microsoeconds (if 'ms' is false)
static void start_timer(U16 time, bit ms, void (* callback)()) {
    Timer16_set_clock(0); // stop timer
    Timer16_set_counter(0); // reset counter
    Timer16_set_compare_a(ms ? (time * 31) : time);
    timer_callback = callback;
    // 2 = prescale 1/8 (1 tick = 1 usec)
    // 4 = prescale 1/256 (1 tick = 32 usec)
    Timer16_set_clock(ms ? 4 : 2); 
}

// start a timer which will reset the state of the system in 500ms (if it isn't cancelled first)
static void start_reset_timer() {
    start_timer(500, TRUE, &reset);
}

// send an 8-bit command to the keyboard
static void send_command(U8 cmd) {
    clock_callback = NULL;
    timer_callback = NULL;
    kbd_data = cmd;
    bit_count = 0;
    disable_clock_interrupt();
    DDRC |= 0x40; // set data line to output mode
    PORTC &= ~0x40; // drive data line low
    enable_clock_interrupt(TRUE, &send_bit);
    start_reset_timer();
}

// tell the keyboard that we are done sending a command, and we are ready to receive the response
static void signal_ready_to_receive() {
    PORTC |= 0x40; // drive data line high
    enable_clock_interrupt(TRUE, &set_receive_mode);
    start_reset_timer();
}
 
// switch from sending mode to receiving mode
static void set_receive_mode() {
    disable_clock_interrupt();
    kbd_data = 0;
    bit_count = 0;
    DDRC &= ~0x40; // set data line to input mode
    PORTC |= 0x40; // enable pullup resistor on data line
    enable_clock_interrupt(FALSE, &receive_bit);
}

// send one bit of kbd_data to the keyboard
static void send_bit() {
    U8 mask;
    
    mask = 1 << (7 - bit_count);
    if (kbd_data & mask) {
        // drive data line high
        PORTC |= 0x40;
    } else {
        // drive data line low
        PORTC &= ~0x40; 
    }
    bit_count++;
    if (bit_count == 8) {
        // sent entire 8-bit command. in 400usec, signal the keyboard that
        // we are ready to receive its response
        disable_clock_interrupt();
        start_timer(400, FALSE, &signal_ready_to_receive);
    }
}

// read one bit from the keyboard into kbd_data
static void receive_bit() {
    U8 mask;
    
    if (PINC & 0x40) {
        // data line is high.
        mask = 1 << (7 - bit_count);
        kbd_data |= mask; 
    }
    bit_count++;
    if (bit_count == 8) {
        // received entire 8-bit response. send a new command
        if (! kbd_connected) {
            // TODO check the model number to make sure it is correct (i.e. not noise)
            kbd_connected = TRUE;
        } else {
            process_response();
        }
        send_command(keypad_mode ? INSTANT_CMD : INQUIRY_CMD);
    }
}
 
// keyboard clock line interrupt handler
ISR(INT4_vect) {
    if (clock_callback != NULL) {
        clock_callback();
    }
}

// timer1 interrupt handler
ISR(TIMER1_COMPA_vect) {
    Timer16_set_clock(0); 
    if (timer_callback != NULL) {
        timer_callback();
    }
}

